@@ -50,8 +50,9 @@
 #   libproc-background-perl libproc-processtable-perl libemail-valid-perl \
 #   libnet-smtps-perl libmail-imapclient-perl libnet-ldap-perl libemail-valid-perl \
 #   libwww-perl liburi-perl
+# optional: libtext-qrcode-perl
 #
-# (c) 2020 Emmanuel Farhi - GRADES - Synchrotron Soleil. AGPL3.
+# (c) 2020- Emmanuel Farhi - GRADES - Synchrotron Soleil. AGPL3.
 # https://gitlab.com/soleil-data-treatment/remote-desktop
 
 
@@ -117,7 +118,7 @@ use constant IS_CGI      => IS_MOD_PERL || exists $ENV{'GATEWAY_INTERFACE'};
 # we use a Hash to store the configuration. This is simpler to pass to functions.
 our %config;
 
-$config{version}                  = "22.04.22";  # year.month.day
+$config{version}                  = "23.02.15";  # year.month.day
 
 require '/etc/qemu-web-desktop/config.pl';
 
@@ -155,7 +156,7 @@ for(my $i = 0; $i < @ARGV; $i++) {
     foreach my $key (keys %config) {
       print STDERR "  --$key=VALUE [$config{$key}]\n";
     }
-    print "\n(c) 2020 Emmanuel Farhi - GRADES - Synchrotron Soleil. AGPL3.\n";
+    print "\n(c) 2020- Emmanuel Farhi - GRADES - Synchrotron Soleil. AGPL3.\n";
     exit;
   } elsif (/^--(\w+)=(\w+)$/) {      # e.g. '--opt=value'
     if (exists($config{$1})) {
@@ -212,6 +213,7 @@ $session{name}        = File::Basename::fileparse($session{dir_snapshot});
 $session{snapshot}    = "$session{dir_snapshot}/$config{service}.qcow2";
 $session{json}        = "$config{dir_cfg}/$session{name}.json";
 $session{output}      = "$session{dir_snapshot}/index.html";
+$session{qrcode}      = "";
 
 $session{user}        = "";
 $session{password}    = "";
@@ -322,9 +324,9 @@ $output .= <<END_HTML;
   <title>$config{service}: $session{machine} [$session{server_name}]</title>
 </head>
 <body>
-  <a href="$config{http}://$session{server_name}/qemu-web-desktop/" target=_blank>
+  <a href="$config{http}://$session{server_name}/$config{service}/" target=_blank>
   <img alt="DARTS" title="DARTS"
-    src="$config{http}://$session{server_name}/qemu-web-desktop/images/darts_logo.png"
+    src="$config{http}://$session{server_name}/$config{service}/images/darts_logo.png"
     align="right" height="128" width="173"></a>
   <h1>Data Analysis Remote Treatment Service: $session{machine}</h1>
   <hr><ul>
@@ -343,7 +345,7 @@ if ($session{session_stop}) {
   }
   print "Content-type:text/html\r\n\r\n";
   print "$output\n\n";
-  print "OK Stopped $session{session_stop} $session{machine} that was started on [$session{date}] by $session{user}\n";
+  print "OK Stopped $session{session_stop} $session{machine} that was started on [$session{date}] by $session{user} $session{user_email}\n";
   exit;
 }
 
@@ -540,9 +542,10 @@ if ($need_fallback_server and length($session{fallback_servers})) {
 # User credentials checks, find communication ports
 # ------------------------------------------------------------------------------
 {
-  (my $out, my $err) = session_authenticate(\%config, \%session);
+  (my $out, my $err, my $session_ref) = session_authenticate(\%config, \%session);
   $output .= $out;
   $error  .= $err;
+  %session     = %{ $session_ref };
 }
 
 if (defined($session{oneshot}) and $session{oneshot} =~ /yes|oneshot|true|1/i) {
@@ -611,7 +614,11 @@ if (not $error) {
     $res      = `$cmd`; # execute command
     $output   .= "<li>$ok Will use ISO from ";
   } else {
-    $cmd      = "qemu-img create -b $config{dir_machines}/$session{machine}"
+    # qemu-img now requires to specify master "backing" file format with '-F FMT'
+    my $img_info = `qemu-img info $config{dir_machines}/$session{machine}`;
+    my @backing_format = $img_info =~ /file format:\s*([a-z0-9]+)/i;
+    my $F     = $backing_format[0];
+    $cmd      = "qemu-img create -F $F -b $config{dir_machines}/$session{machine}"
               . " -f qcow2 $session{snapshot}";
     $res      = `$cmd`; # execute command
     $output  .= "<li>$ok Creating snapshot from ";
@@ -682,7 +689,7 @@ if (not $error) {
   for(my $i = 0; $i <= $#mounts; $i++) {
     if (-d $mounts[$i]) { # mount must exist as a directory
       my $tag = (split '/', $mounts[$i])[-1];
-      $cmd .= " -fsdev local,security_model=passthrough,id=fsdev$i,path=$mounts[$i] -device virtio-9p-pci,id=fs$i,fsdev=fsdev$i,mount_tag=host_$tag";
+      $cmd .= " -fsdev local,security_model=none,id=fsdev$i,path=$mounts[$i] -device virtio-9p-pci,id=fs$i,fsdev=fsdev$i,mount_tag=host_$tag";
     }
   }
   
@@ -755,14 +762,6 @@ if (not $error) {
       " --token-plugin=TokenFile" .
       " --token-source=$dir_token";
   }
-
-# unfortunately, using a Token does not seem to work
-  
-  if ($config{service_port_multiple}) {
-    $cmd .= " --record=$session{dir_snapshot}/websocket.log";
-  } else {
-    $cmd .= " --record=$config{dir_snapshots}/websocket.log";
-  }
   
   if ($session{oneshot}) { $cmd .= " --run-once"; }
   
@@ -832,12 +831,43 @@ if (not $error) {
   my $output_share = "";
   my $output_persistent = "";
   if (not $session{oneshot}) {
-    $output_share = "<button onClick=\"SelfCopy(this.id)\" id=\"$session{url} $session{token_vnc}\" style=\"padding:0;border:0;background:none;outline:none\"><img alt=\"copy\" src=\"$config{http}://$session{server_name}/qemu-web-desktop/images/share-button.png\" title=\"Copy link to clipboard, share it with your colleagues.\" height=40></button>";
+    $output_share = "<button onClick=\"SelfCopy(this.id)\" id=\"$session{url} $session{token_vnc}\" style=\"padding:0;border:0;background:none;outline:none\"><img alt=\"copy\" src=\"$config{http}://$session{server_name}/$config{service}/images/share-button.png\" title=\"Copy link to clipboard, share it with your colleagues.\" height=40></button>";
     $output_persistent = "<li>You can close the browser and reconnect any time (within life-time) with the link above.</li>" .
     "<li>Select the <span style=\"color:green\">[Manage sessions]</span> item in the service login page to list, reconnect or abort your sessions.</li>" .
     "<li>You can collaborate in the same session with your colleagues. Just send them the link above.</li>" .
     "<li>Please <b>shut-down the machine properly</b> (do not just logout or suspend).</li>";
   }
+  
+  # Conditional import of Text::QRCode module when installed
+  my $icon = "
+        <img alt='$session{machine}' title='$session{machine}'
+        src='$config{http}://$session{server_name}/$config{service}/images/logo-system.png'
+        align='center' border='1' height='128'>";
+  
+  my $qr_enabled = eval
+  {
+    require Text::QRCode;
+    Text::QRCode->import();
+    1;
+  };
+  
+  if ($qr_enabled) {
+    # https://metacpan.org/pod/Text::QRCode
+    my $arrayref = Text::QRCode->new()->plot("$session{url}");
+    $session{qrcode} = join "\n", map { join '', @$_ } @$arrayref; 
+  
+    # code from HTML::QRCode
+    my $w = "<td style=\"border:0;margin:0;padding:0;width:3px;height:3px;background-color: white;\">";
+    my $b = "<td style=\"border:0;margin:0;padding:0;width:3px;height:3px;background-color: black;\">";
+ 
+    $icon
+        = '<table style="margin:0;padding:0;border-width:0;border-spacing:0;">';
+    $icon
+        .= '<tr style="border:0;margin:0;padding:0;">'
+        . join( '', map { $_ eq '*' ? $b : $w } @$_ ) . '</tr>'
+        for (@$arrayref);
+    $icon .= '</table>';
+  }
 
   $output .= <<END_HTML;
     <h1>Hello $session{user} !</h1>
@@ -846,14 +876,11 @@ if (not $error) {
     Click on the following link.</p>
     
     <div style="text-align: center;">
-      <a href=$session{url} target=_blank>
-        <img alt="$session{machine}" title="$session{machine}"
-        src="$config{http}://$session{server_name}/qemu-web-desktop/images/logo-system.png"
-        align="center" border="1" height="128">
-      </a><br></br>
+      <a href=$session{url} target=_blank>$icon</a>
+      <br></br>
       <b><a href="$session{url}" target=_blank>
       <img alt="connect" 
-        src="$config{http}://$session{server_name}/qemu-web-desktop/images/connect-button.jpeg" 
+        src="$config{http}://$session{server_name}/$config{service}/images/connect-button.jpeg" 
         title="Connect !" height=40></a></b> 
         $output_share
         $output_token
@@ -867,15 +894,15 @@ if (not $error) {
       (e.g. mounted disk, ssh/sftp, Dropbox, OwnCloud...).</li>
     <li>To <b>kill</b> this session, click on <form action="$config{http}://$session{server_name}/cgi-bin/qemu-web-desktop.pl" method=post target=_blank>
     <input type=hidden name=session_stop value="$session{name}">
-    <input type=image alt="stop" src="$config{http}://$session{server_name}/qemu-web-desktop/images/stop-button.png" title=STOP height="32">
+    <input type=image alt="stop" src="$config{http}://$session{server_name}/$config{service}/images/stop-button.png" title=STOP height="32">
     </form></li>
     </ul></p>
 
     <hr>
-    <small><a href="$config{http}://$session{server_name}/qemu-web-desktop/" target=_blank>Data Analysis Remote Treatment Service</a> (c) 2020 - <a href="http://www.synchrotron-soleil.fr" target="_top">Synchrotron Soleil</a> - Thanks for using our data analysis services !</small>
+    <small><a href="$config{http}://$session{server_name}/$config{service}/" target=_blank>Data Analysis Remote Treatment Service</a> (c) 2020- - <a href="http://www.synchrotron-soleil.fr" target="_top">Synchrotron Soleil</a> - Thanks for using our data analysis services !</small>
     <a href="http://www.synchrotron-soleil.fr" target="_top">
     <img alt="SOLEIL" title="SOLEIL"
-    src="$config{http}://$session{server_name}/qemu-web-desktop/images/logo_soleil.png"
+    src="$config{http}://$session{server_name}/$config{service}/images/logo_soleil.png"
     align="right" border="0" height="48"></a>
     </body>
     
@@ -908,7 +935,7 @@ if (not $error) {
   print $fh "$output\n";
   close $fh;
   
-  print STDERR $0.": $session{name}: $session{date}: START $session{machine} by $session{user}\n";
+  print STDERR $0.": $session{name}: $session{date}: START $session{machine} by $session{user} $session{user_email}\n";
   print STDERR $0.": $session{name}: json:  $session{json}\n";
   print STDERR $0.": $session{name}: URL:   $session{url}\n";
   print STDERR $0.": $session{name}: PIDs:  @{$session{pid}}\n" if ($session{pid});
@@ -1024,10 +1051,12 @@ sub session_stop {
     { unlink($session{output}); }
   if ($session{token_file} and -e $session{token_file})
     { unlink($session{token_file}); }
+  if ($session{qrcode} and -e $session{qrcode})
+    { unlink($session{qrcode}); }
   
   my $now         = localtime();
   if ($session{remote_host}) { 
-    print STDERR $0." [$now] STOP $session{name} $session{machine} started on [$session{date}] for $session{user}\@$session{remote_host}\n";
+    print STDERR $0." [$now] STOP $session{name} $session{machine} started on [$session{date}] for $session{user}\@$session{remote_host} $session{user_email}\n";
   }
   
   # make sure QEMU/noVNC and asssigned SHELLs are killed
@@ -1043,12 +1072,60 @@ sub session_stop {
   
 } # session_stop
 
+# (TF, $content) = session_test_script(\%session, content)
+# Test the script '$content' after replacing symbols
+#   The test expression must be found with syntax 'if(exp):'. It is extracted,
+#   removed from the $content (which is returned after removal). Then it is 
+#   evaluated, and its result is placed in TF (returned). False is returned
+#   when the test can not be evaluated (fails).
+# return (TF, $content)
+sub session_test_script {
+  my $session_ref = shift;
+  my $content     = shift;
+  my $TF          = 0;
+  
+  if (not $session_ref) { return (0,$content); }
+  if (not $content)     { return (0,$content); }
+  
+  my %session = %{ $session_ref };
+  
+  # replace @symbols@ (globally)
+  $content=~s/\@USER\@/$session{user}/g;
+  $content=~s/\@SESSION_NAME\@/$session{name}/g;
+  $content=~s/\@VM\@/$session{machine}/g;
+  $content=~s/\@PW\@/$session{password}/g;
+  
+  # trim
+  $content =~ s/^\s+|\s+$//g ; 
+
+  # now search for the expression "if(EXPR):". Get 1st match.
+  if ($content =~ m/if\((.*)\):/) {
+    my $test = $1;
+    $content =~ s/if\((.*)\)://; # remove match
+    $content =~ s/^\s+|\s+$//g ; # trim again
+    # evaluate the EXPR in test.
+    $TF = eval $test;
+    if ($@) { $TF = 0; } # error -> fails
+    return ($TF,$content);
+  } else {
+    # no match: use content as is.
+    return (1, $content);
+  }
+  
+} # session_test_script
+
 # session_run_script(\%config, \%session, $script): install a script for boot in VM
 # The script can be specifyed as:
 #   "http://some/url"
 #   "/some/local/path/to/script"
 #   "exec: some commands" separated with EOL (\n) or ';'
 #   "virt-customize: commands" separated with EOL (\n)
+#
+# In addition, the script text (expression, file or URL) may contain the syntax:
+#   if(EXPR):
+# so that the script content is evaluated when EXPR is true.
+# The script content and EXPR may use the @USER@ @SESSION_NAME@ and @VM@ symbols,  
+# which are replaced by the user name, the session name and the VM name.
 sub session_run_script {
   my $config_ref  = shift;
   my $session_ref = shift;
@@ -1059,9 +1136,12 @@ sub session_run_script {
   my %config  = %{ $config_ref };
   my %session = %{ $session_ref };
   
-  if (length($script) > $config{service_max_script_length}) {
+  if (length($script) > $config{service_max_script_length} || not $script) {
     return "";
   }
+  my $TF=0;
+  ($TF,$script) = session_test_script(\%session, $script);
+  if (not $TF) { return ""; }
   
   my $content     = undef;
   my $virt_customize_commands = 0;
@@ -1099,11 +1179,10 @@ sub session_run_script {
     # remove invalid characters
     $content=~s/[^\x00-\x7f]//g;
     if (length($content) < $config{service_max_script_length}) {
-
-      # replace @symbols@ (e.g. for auto login)
-      $content=~s/\@USER\@/$session{user}/g;
-      $content=~s/\@SESSION_NAME\@/$session{name}/g;
-      $content=~s/\@VM\@/$session{machine}/g;
+    
+      # test if script is active
+      ($TF, $content) = session_test_script(\%session, $content);
+      if (not $TF) { return ""; }
       
       # store script in our 'snapshot' directory for local exec.
       open my $fh, ">", "$session{dir_snapshot}/configuration_script";
@@ -1205,7 +1284,7 @@ sub service_monitor {
   # display a table with current info
   $out .= "</ul><br><br><h1>Current $config{service} service status ($now)</h1>";
   $out .= "This page lists the service status, and your active sessions. <br>";
-  $out .= "<img src=\"$config{http}://$config{server_name}/qemu-web-desktop/images/virtualmachines.png\" height=128 align=right>";
+  $out .= "<img src=\"$config{http}://$config{server_name}/$config{service}/images/virtualmachines.png\" height=128 align=right>";
   $out .= "<table  border='1'>\n";
   $out .= "<tr><th>Server           </th><th>$config{server_name} running '$config{service}' version $config{version}</th></tr>\n";
   $out .= "<tr><td>#CPU total       </td><td>$cpu_count</td></tr>\n";
@@ -1253,8 +1332,12 @@ sub service_monitor {
               $out .= "<tr>";
               $out .= "<td>$session{date}</td>";
               $out .= "<td>$session{name}</td>";
-              $out .= "<td>$session{machine}</td>";
-              $out .= "<td>$session{user}</td>";
+              $out .= "<td><a href='$config{http}://$session{server_name}/$config{service}/machines/$session{machine}'>$session{machine}</a></td>";
+              if ($session{user_email}) {
+                $out .= "<td><a href='mailto:$session{user_email}'>$session{user}</a></td>";
+              } else  {
+                $out .= "<td>$session{user}</td>";
+              }
               $out .= "<td>$session{cpu}</td>";
               $out .= "<td>$session{memory}</td>";
               if ($session{gpu}) {
@@ -1266,7 +1349,7 @@ sub service_monitor {
                 $out .= "<td>one-shot";
               } else {
                 $out .= "<td><a href='$session{url}' target=_blank><img alt=connect 
-          src=\"$config{http}://$session{server_name}/qemu-web-desktop/images/connect-button.jpeg\" 
+          src=\"$config{http}://$session{server_name}/$config{service}/images/connect-button.jpeg\" 
           title=Connect height=32></a>";
                 if ($session{token_vnc} && $config{service_port_multiple}) {
                   $out .= " $session{token_vnc}";
@@ -1274,7 +1357,7 @@ sub service_monitor {
               }
               $out .= "</td>";
               # create a STOP button for each
-              $out .= "<td>@pids <form action=\"$config{http}://$session{server_name}/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_stop value=\"$session{name}\"><input type=image alt=stop src=\"$config{http}://$session{server_name}/qemu-web-desktop/images/stop-button.png\" title=STOP height=32></form></td>\n";
+              $out .= "<td>@pids <form action=\"$config{http}://$session{server_name}/cgi-bin/qemu-web-desktop.pl\" method=post target=_blank><input type=hidden name=session_stop value=\"$session{name}\"><input type=image alt=stop src=\"$config{http}://$session{server_name}/$config{service}/images/stop-button.png\" title=STOP height=32></form></td>\n";
             }
           }
         }
@@ -1395,7 +1478,7 @@ sub service_check_user_max {
 
 # session_authenticate(\%config, \%session)
 #   check user credentials with SMTP, LDAP, IMAP and sendemail.
-#   return: ($out, $err)
+#   return: ($out, $err, $session)
 sub session_authenticate {
 
   my $config_ref  = shift;
@@ -1429,7 +1512,16 @@ sub session_authenticate {
       $authenticated .= session_check_smtp(\%config, \%session); # checks SMTP("user","password")
     }
     if (index($authenticated, "SUCCESS") < 0 and $config{check_user_with_ldap}) {
-      $authenticated .= session_check_ldap(\%config, \%session); # checks LDAP("user","password")
+      my $authenticated_ldap = "";
+      my $session_ref;
+      ($authenticated_ldap, $session_ref) = session_check_ldap(\%config, \%session); # checks LDAP("user","password")
+      $authenticated .= $authenticated_ldap;
+      if ($session_ref) {
+        %session     = %{ $session_ref };
+      }
+    }
+    if ($config{check_user_custom}) {
+      $authenticated = session_check_custom(\%config, \%session, $authenticated); # checks custom("user","password",$authenticated)
     }
     # now we search for a "SUCCESS"
     if (index($authenticated, "SUCCESS") > -1) {
@@ -1438,12 +1530,12 @@ sub session_authenticate {
       $out .= "<li><b><font color=orange>[WARNING]</font></b> Service is running without user authentication.</li>\n";
       # no authentication configured...
     } else {
-      $err  .= "User $session{user} failed authentication. Check your username / password:  $authenticated."; 
+      $err  .= "User $session{user} failed authentication. Check your username / password:  $authenticated"; 
     }
     
     
   } # authentication block
-  return ($out, $err);
+  return ($out, $err, \%session);
   
 } # session_authenticate
 
@@ -1531,7 +1623,7 @@ sub session_check_imap {
 } # session_check_imap
 
 # session_check_ldap(\%config, \%session)
-#   ldap_server is needed.
+#   ldap_server is needed. Return ($res, $session)
 #   return ""         when no check is done
 #          "FAILED"   when authentication failed
 #          "SUCCESS"  when authentication succeeded
@@ -1599,10 +1691,55 @@ sub session_check_ldap {
   }
   
   $ldap->unbind;
-  return $res;
+  return ($res,\%session);
 
 } # session_check_ldap
 
+# session_check_custom(\%config, \%session, $authenticated)
+#   check with a user function defined in %config
+#   return ""         when no check is done
+#          "FAILED"   when authentication failed
+#          "SUCCESS"  when authentication succeeded
+#
+# specify an authenticator function in /etc/qemu-web-desktop/config.pl such as:
+#
+#  sub check_user_func {
+#    my $user = shift;
+#    my $pw   = shift;
+#    my $auth = shift; # previous auth
+#    return "$auth and SUCCESS: [Custom] $user authenticated.";
+#  }
+#
+#  $config{check_user_custom}   = \&check_user_func;
+
+sub session_check_custom {
+  my $config_ref  = shift;
+  my $session_ref = shift;
+  my $authenticated = shift;
+
+  if (not $config_ref or not $session_ref) { return ""; }
+  my %config      = %{ $config_ref };
+  my %session     = %{ $session_ref };
+
+  if (not %config or not %session) { return ""; }
+  
+  # check if existing custom auth
+  if (not $config{check_user_custom}) { return ""; }
+  my $res    = "";
+  
+  if (not $session{user}) {
+    return "FAILED: [custom auth] Missing Username.";
+  }
+  
+  # execute the referenced function
+  my $cref = $config{check_user_custom};
+  eval { $res = &$cref($session{user}, $session{password}, $authenticated, \%session); }; 
+  print STDERR "$0: Error when executing custom authenticator: $@" if $@;
+  
+  return $res;
+
+} # session_check_custom
+
 # session_use_gpu(\%config, $pci): return true if the given PCI is used by any running session
 sub session_use_gpu {
   my $config_ref  = shift;
