@@ -13,7 +13,7 @@ $config{service}                  = "qemu-web-desktop";
 
 # full path to Apache HTML root
 #   Apache default on Debian is /var/www/html
-$config{dir_html}                 = "/usr/share/qemu-web-desktop/html";
+$config{dir_html}                 = "/usr/share/$config{service}/html";
 
 # full path to root of the service area
 #   e.g. /var/www/html/desktop
@@ -122,7 +122,12 @@ $config{service_max_script_length}= 65535; # e.g. 65535
 #   "virt-customize: commands" separated with EOL
 # The symbols `@USER@` `@SESSION_NAME@` and `@VM@` are replaced by the user 
 # name, the session ID, and the virtual machine name. 
-# They are executed at boot, in background, as root.
+#
+# In addition, when the above script description is preceded by `if(EXPR):` 
+# the given expression is evaluated and the script is only executed with Perl
+# when result is True. The `EXPR` condition may use the `@...@` symbols above.
+#
+# Scripts are executed at boot, in background, as root.
 # It is possible to define as many scripts as needed. Strings are separated by ','
 my @config_scripts=("");
 $config{config_script}            = [@config_scripts];
@@ -187,6 +192,39 @@ $config{check_user_with_imap}     = 0;
 $config{check_user_with_smtp}     = 0;
 $config{check_user_with_ldap}     = 0;
 
+# Custom user authentication.
+#   When defined, this authenticator is ALWAYS executed (whatever be the other 
+#   authenticator results).
+#
+# First, specify a function that should get (user, pw, previous_auth) as 
+# arguments and return a string starting by "SUCCESS" or "FAILED". The default  
+# return value should be the previous authenticator results.
+# Any "SUCCESS" in the returned string fully qualifies the authentication.
+#
+#  sub check_user_func {
+#    my $user          = shift;
+#    my $pw            = shift;
+#    my $authenticated = shift; # previous authenticator results
+#    my $res           = "";
+#
+#    res = "SUCCESS: [Custom] $user authenticated.";
+#
+#    res = "FAILED: [Custom] $user failed authentication.";
+#
+#    return "$authenticated and $res";
+#  }
+#
+# Then send its reference to the configuration:
+#
+#   $config{check_user_custom} = \&check_user_func;
+#
+# or directly as an anonymous function
+#
+#   $config{check_user_custom} = sub { };
+
+$config{check_user_custom}   = "";
+
+
 # Encryption (HTTPS). Set these to valid 'crt' and 'key' files with proper certificates
 # to allow secured encryption of websockify/novnc. The CRT and KEY should both be set.
 # OpenStack DEFAULT.key DEFAULT.cert
